import os
import typing
from pathlib import Path

import pandas as pd

from datamazing.pandas import io
from datamazing.pandas.types import TimeInterval


def get_filepath(filename: str, subfolder: str = "data"):
    return Path(os.environ["PYTEST_CURRENT_TEST"]).parent / subfolder / filename


def make_df(data: list[dict]):
    return pd.DataFrame.from_records(data)


def make_series(data: dict):
    return pd.Series(data)


def read_df(filename: str, subfolder: str = "data") -> pd.DataFrame:
    """
    Read pandas DataFrame from test data.
    Datetimes and timedeltas are inferred automatically.

    Args:
        filename (str): Name of CSV file with test data
        subfolder (str, optional): Subfolder relative to test being
            run currently (taken from  the environment variable PYTEST_CURRENT_TEST),
            from where to read the test data. Defaults to "data".

    Example:
    >>> read_df(filename="data.df.csv")
    """
    filepath = get_filepath(filename, subfolder)
    df = io.read_csv(filepath)

    return df


class TestDatabase:
    def __init__(
        self, file_pattern: str, subfolder: str = "data", time_column: str = "time_utc"
    ):
        """Read Database from test data.
        Datetimes and timedeltas are inferred automatically.

        Args:
            file_pattern (str): Pattern of CSV files with test data
                using placeholder `{table_name}`
            subfolder (str, optional): Subfolder relative to test being
                run currently (taken from  the environment variable
                PYTEST_CURRENT_TEST), from where to read the test data.
                Defaults to "data".

        Example:
        >>> TestData(file_pattern="{table_name}.df.csv")
        """
        self.file_pattern = file_pattern
        self.subfolder = subfolder
        self.time_column = time_column

    def query(
        self,
        table_name: str,
        time_interval: typing.Optional[TimeInterval] = None,
        filters: typing.Optional[dict[str, object]] = None,
    ) -> pd.DataFrame:
        # get filename associated with table
        filename = self.file_pattern.format(table_name=table_name)

        df = read_df(filename, subfolder=self.subfolder)

        if time_interval:
            # filter to time interval
            is_after_start = df[self.time_column] >= time_interval.left
            is_before_end = df[self.time_column] <= time_interval.right
            is_in_interval = is_after_start & is_before_end
            df = df[is_in_interval]

        if filters:
            for key, value in filters.items():
                if key not in df.columns:
                    raise KeyError(f"The column {key} is not found in {table_name}.")
                if not isinstance(value, list):
                    value = [value]
                df = df[df[key].isin(value)]

            df = df.reset_index(drop=True)

        return df



from functools import reduce

import numpy as np
import pandas as pd

# we dont use pd.Timestamp.min and pd.Timestamp.max,
# as these values are too obscure
MIN_TIMESTAMP = pd.Timestamp("1900-01-01T00:00:00")
MAX_TIMESTAMP = pd.Timestamp("2200-01-01T00:00:00")


def fill_empty_periods(df, period: tuple[str, str]):
    df = df.copy()
    min_timestamp = MIN_TIMESTAMP.tz_localize(getattr(df.dtypes[period[0]], "tz", None))
    max_timestamp = MAX_TIMESTAMP.tz_localize(getattr(df.dtypes[period[1]], "tz", None))
    df[period[0]] = df[period[0]].fillna(min_timestamp)
    df[period[1]] = df[period[1]].fillna(max_timestamp)
    return df


def latest(
    df: pd.DataFrame,
    on: str,
):
    s = df.sort_values(by=on).iloc[-1]
    s.name = None
    return s


def earliest(df: pd.DataFrame, on: str):
    s = df.sort_values(by=on).iloc[0]
    s.name = None
    return s


def as_of_time(
    df: pd.DataFrame,
    period: tuple[str, str],
    at: pd.Timestamp,
    closed: str = "left",
):
    """Get snapshot of periodized DataFrame
    at a specified timestamp

    Args:
        df (pd.DataFrame): DataFrame
        period (tuple[str, str]): Periodization columns (valid-from, valid-to)
        at (pd.Timestamp): Timestamp to get snapshot at
        closed (str, optional): Boundary of periodization intervals. Defaults to "left".
    """
    df = fill_empty_periods(df, period)

    return df[
        pd.IntervalIndex.from_arrays(
            df[period[0]], df[period[1]], closed=closed
        ).contains(at)
    ]


def shift_time(
    df: pd.DataFrame,
    on: str,
    period: pd.Timedelta,
):
    df = df.copy()
    df[on] = df[on] + period
    return df


def unique_one(s: pd.Series) -> object:
    if len(s.unique()) > 1:
        raise ValueError("Series has more than one unique value")
    return s.iloc[0]


def concat_by_name(dfs: list[pd.DataFrame]) -> pd.DataFrame:
    columns = [df.columns for df in dfs]
    common_cols = reduce(lambda cols1, cols2: cols1.union(cols2), columns).unique()
    return pd.concat([df.reindex(columns=common_cols) for df in dfs], ignore_index=True)


def unpivot(
    df: pd.DataFrame, on: list[str], variable: str = "variable", value: str = "value"
):
    remaining_cols = df.columns.drop(on).tolist()
    df = df.set_index(remaining_cols)
    df.columns.name = variable
    s = df.stack()
    s.name = value
    df = s.reset_index()
    return df


def switch(
    cases: list[tuple[pd.Series, pd.Series | object]],
    default: pd.Series | object,
) -> pd.Series:
    """Choose among a set of values depending on a set of conditions

    Args:
        cases (list[tuple[pd.Series, pd.Series  |  object]]):
            List of cases, with each element being a tuple
            of the form (when, then). I.e. if `when` is satisfied,
            choose `then`. If multiple cases are satisfied,
            the first one is choosen.
        default (pd.Series | object):
            If no cases are satisfied, this value will be choosen.

    Example:
    >>> df["sign"] = pdz.switch(
    ...     cases=[
    ...         (df["value"] > 0, "positive"),
    ...         (df["value"] < 0, "negative"),
    ...     ],
    ...     default="zero",
    ... )
    """
    array = np.select(
        condlist=[case[0] for case in cases],
        choicelist=[case[1] for case in cases],
        default=default,
    )

    return pd.Series(array)



import pandas as pd

ZERO_INTERVAL = pd.Timedelta(microseconds=0)


class Resampler:
    def __init__(
        self,
        df: pd.DataFrame,
        on: str,
        resolution: pd.Timedelta,
    ):
        self.df = df
        self.on = on
        self.resolution = resolution

    def agg(self, method: str | dict, edge: str = "left"):
        """Aggregate (downsample) time series.
        For example, if the input is a time series H with
        hourly resolution, then we can aggregate, using
        the mean, to daily resolution, and produce a new
        time series D. In other words:

            - Input:    H(hour)
            - Output:   D(day) = mean( H(hour) for hour in day)

        Args:
            method (str | dict): Name of method (e.g. "sum") or
                dictionary of method(s) to use for each column
                (e.g. {"col1": "mean", "col2": ["min", "max"]})
            edge (str, optional): Which side of the interval
                to use as label ("left" or "right").
                Defaults to "left".
        """
        df = (
            self.df.sort_values(by=self.on)
            .set_index(keys=self.on)
            .resample(rule=self.resolution, label=edge, closed=edge)
            .aggregate(method, numeric_only=True)
            .reset_index()
        )

        if isinstance(df.columns, pd.MultiIndex):
            df.columns = ["_".join(col).strip("_") for col in df.columns]

        return df

    def interpolate(
        self,
        method: str = "linear",
        order=None,
    ):
        """Interpolate (upsample) time series.
        For example, if the input is a time series D with
        daily resolution, then we can interpolate, using
        a linear function, to hourly resolution, and
        produce a new time series H. In other words:

            - Input:    D(day)
            - Output:   H(hour) = interpolate(hour between day+0 and day+1)

        Args:
            method (str): Interpolation method. Must be one of
                'linear'
                'ffill'
                'bfill'
                'nearest'
                'zero'
                'slinear'
                'quadratic'
                'cubic'
                'barycentric'
                'polynomial'
                'krogh'
                'piecewise_polynomial'
                'spline'
                'pchip'
                'akima'
                'cubicspline'

            order (int): Order of interpolation method when using one of
                'polynomial'
                'spline'

        Raises:
            ValueError: Currently, interpolation can only be done
                when the original time series has a resolution
                which is an integer factor of the upsample resolution.
                Thus, one can go from 1 hour-resolution to 5 minute
                -resolution, but not the other way around. This
                restriction is purely a pandas issue, and can be
                fixed if desired.
        """
        # only allow certain interpolation methods
        # to avoid untested behavior
        supported_methods = [
            "linear",
            "ffill",
            "bfill",
            "nearest",
            "zero",
            "slinear",
            "quadratic",
            "cubic",
            "barycentric",
            "polynomial",
            "krogh",
            "piecewise_polynomial",
            "spline",
            "pchip",
            "akima",
            "cubicspline",
        ]
        if method not in supported_methods:
            raise ValueError(f"Method must be one of {supported_methods}")

        # in pandas, the interpolation method `linear`
        # assumes that the timestamps are evenly spaced.
        # This might no be the case when reindexing, so
        # instead we can use the method `time`
        if method == "linear":
            method = "time"

        # pandas doesn't handle empty dataframes very well
        if self.df.empty:
            return self.df

        df = self.df.sort_values(by=self.on)

        diff_factors = df[self.on].diff().dropna() / self.resolution

        if ((diff_factors % 1) != 0).any():
            raise ValueError(
                f"Interpolation only possible "
                f"when the original time resolution "
                f"is an integer factor of the upsample "
                f"resolution '{self.resolution}'."
            )

        start_time = df[self.on].min()
        end_time = df[self.on].max()

        # use correct limit direction
        if method in ["bfill", "backfill"]:
            limit_direction = "backward"
        elif method in ["pad", "ffill"]:
            limit_direction = "forward"
        else:
            limit_direction = None

        if order:
            interpolation_options = {"order": order}
        else:
            interpolation_options = {}

        df = (
            df.set_index(self.on)
            .resample(rule=self.resolution)
            .interpolate(
                method, limit_direction=limit_direction, **interpolation_options
            )
            .reset_index()
        )

        # after resampling, pandas might leave
        # timestamps outside of the original interval
        if not df.empty:
            df = df[df[self.on].between(start_time, end_time)]

        return df

    def granulate(self, block_size: pd.Timedelta, edge: str = "left"):
        """Fine-grain (upsample) time series.
        For example, if the input is a time series D with
        daily resolution, then we can granulate to hourly
        resolution, and produce a new time series H.
        In other words:

            - Input:    D(day)
            - Output:   H(hour) = D(day containing hour)


        Args:
            block_size: The block size in which the time is granulated.
            edge (str, optional): Which side of the interval
                to use as label ("left" or "right").
                Defaults to "left".
        """
        # pandas doesn't handle empty dataframes very well
        if self.df.empty:
            return self.df

        df = self.df.sort_values(by=self.on)

        if edge == "left":
            left_margin = ZERO_INTERVAL
            right_margin = block_size
        elif edge == "right":
            left_margin = -block_size
            right_margin = ZERO_INTERVAL
        else:
            raise ValueError(f"Unsupported value {edge} for `edge`")

        df[self.on] = df.apply(
            lambda row: pd.date_range(
                row[self.on] + left_margin,
                row[self.on] + right_margin,
                freq=self.resolution,
                inclusive=edge,
            ),
            axis="columns",
        )

        df = df.explode(column=self.on).reset_index(drop=True)

        return df


def resample(df: pd.DataFrame, on: str, resolution: pd.Timedelta):
    return Resampler(df, on, resolution)



def _list(a):
    """Convert element to list of not already"""
    if isinstance(a, (tuple, set)):
        return list(a)
    if not isinstance(a, list):
        return [a]
    return a


def _concat(a, b):
    """Concatenate elements to list"""
    return _list(a) + _list(b)



from functools import reduce

import pandas as pd

from datamazing._conform import _concat
from datamazing.pandas.transformations import basic


def merge_many(
    dfs: list[pd.DataFrame],
    on: list[str],
    how: str = "inner",
) -> pd.DataFrame:
    return reduce(lambda df1, df2: df1.merge(df2, on=on, how=how), dfs)


def merge(
    left: pd.DataFrame,
    right: pd.DataFrame,
    left_on: list[str] | None = None,
    right_on: list[str] | None = None,
    on: list[str] | None = None,
    left_time: str | None = None,
    right_period: tuple[str, str] | None = None,
    how: str = "inner",
    suffixes: tuple[str, str] = ("_x", "_y"),
):
    if on:
        if left_on is not None or right_on is not None:
            # raise same error as pd.merge
            raise pd.errors.MergeError(
                'Can only pass argument "on" OR "left_on" '
                'and "right_on", not a combination of both.'
            )
        left_on = on
        right_on = on

    if bool(left_time) != bool(right_period):
        raise ValueError(
            "Both `left_time` and `right_period` must be set when time-merging"
        )

    if left_time and right_period:
        if how not in ["inner", "left"]:
            raise ValueError(
                'Only "inner" or "left" join is supported when time-merging'
            )

        # sort values in order to use `pd.merge_asof`
        left = left.sort_values(by=left_time)

        right = basic.fill_empty_periods(df=right, period=right_period)
        right = right.sort_values(by=right_period[0])

        # make left-join, matching the latest right-start-time
        # which is below the left-time
        df = pd.merge_asof(
            left,
            right,
            left_on=left_time,
            right_on=right_period[0],
            left_by=left_on,
            right_by=right_on,
            suffixes=suffixes,
        )

        # remove matches, where the right-end-time is also below the left-time.
        df = df[df[left_time] < df[right_period[1]]]

        # the above code, actually does an inner-join, so we need to
        # add the remaining rows from the left dataframe.
        # we need to take into account overlapping columns,
        # which needs to be appended with the given suffixes
        if how == "left":
            df_right_cols = df.filter(
                list(right.columns) + list(right.columns + suffixes[1]) + [left_time]
            )

            df_right_cols.columns = df_right_cols.columns.str.replace(suffixes[1], "")

            df = pd.merge(
                left,
                df_right_cols,
                left_on=_concat(left_on, left_time),
                right_on=_concat(right_on, left_time),
                how="left",
                suffixes=suffixes,
            )

    else:
        df = pd.merge(
            left, right, left_on=left_on, right_on=right_on, how=how, suffixes=suffixes
        )

    return df


def align(
    left: pd.DataFrame,
    right: pd.DataFrame,
    left_on: list[str] | None = None,
    right_on: list[str] | None = None,
    on: list[str] | None = None,
    how: str = "inner",
):
    if on:
        if left_on is not None or right_on is not None:
            # raise same error as pd.merge
            raise pd.errors.MergeError(
                'Can only pass argument "on" OR "left_on" '
                'and "right_on", not a combination of both.'
            )
        left_on = on
        right_on = on

    left = left.set_index(left_on)
    right = right.set_index(right_on)

    aligned_left, aligned_right = left.align(right, join=how, axis="index")

    aligned_left = aligned_left.reset_index()
    aligned_right = aligned_right.reset_index()

    return aligned_left, aligned_right



from dataclasses import dataclass

import pandas as pd
from pandas.tseries.offsets import DateOffset


@dataclass
class TimeInterval:
    """
    Class representing a time interval

    Args:
        left (pd.Timestamp): Left end of interval
        right (pd.Timestamp): Right end of interval
    """

    left: pd.Timestamp
    right: pd.Timestamp

    def __post_init__(self):
        self.left = pd.Timestamp(self.left)
        self.right = pd.Timestamp(self.right)

    def shift(self, period: pd.Timedelta) -> "TimeInterval":
        return TimeInterval(
            left=self.left + period,
            right=self.right + period,
        )

    def to_range(self, freq: pd.Timedelta, name: str = "time_utc") -> pd.DatetimeIndex:
        return pd.date_range(self.left, self.right, freq=freq, name=name)

    def isoformat(self) -> str:
        return f"{self.left.isoformat()}/{self.right.isoformat()}"

    def tz_localize(self, tz: str) -> "TimeInterval":
        """Localize a TimeInterval to a timezone.

        Args:
            tz (str): The timezone to localize to.

        Returns:
            TimeInterval: The localized TimeInterval.
        """
        return TimeInterval(
            left=self.left.tz_localize(tz),
            right=self.right.tz_localize(tz),
        )

    def tz_convert(self, tz: str) -> "TimeInterval":
        """Convert a TimeInterval to a different timezone.

        Args:
            tz (str): The timezone to convert to.

        Returns:
            TimeInterval: The converted TimeInterval.
        """
        return TimeInterval(
            left=self.left.tz_convert(tz),
            right=self.right.tz_convert(tz),
        )

    @classmethod
    def from_date(cls, date: pd.Timestamp) -> "TimeInterval":
        """Create a TimeInterval from a date. The TimeInterval covers the whole day.
        Can be used to create a TimeInterval for a Danish day.

        Example:
        ```python
        date = pd.Timestamp.utcnow()
        date_cet = date.tz_convert("Europe/Copenhagen")
        time_interval_danish_day_cet = TimeInterval.from_date(date_cet)
        time_interval_danish_day_utc = time_interval_danish_day_cet.tz_convert("UTC")

        >>> time_interval_danish_day_utc
        TimeInterval(
            left=Timestamp('2024-09-26 22:00:00+0000', tz='UTC'),
            right=Timestamp('2024-09-27 22:00:00+0000', tz='UTC')
        )
        ```

        Args:
            date (pd.Timestamp): The date to change to a TimeInterval

        Returns:
            TimeInterval: The TimeInterval from start of day to start of next day
        """
        whole_date = date.floor("D")
        left = whole_date
        right = whole_date + DateOffset(days=1)
        return cls(
            left=left,
            right=right,
        )

    @classmethod
    def fromisoformat(cls, string: str) -> "TimeInterval":
        start, end = string.split("/")
        return cls(
            left=pd.Timestamp(start),
            right=pd.Timestamp(end),
        )

    def __hash__(self):
        return hash((self.left, self.right))



from abc import ABC, abstractmethod

import pandas as pd

from datamazing.pandas.types import TimeInterval


class Database(ABC):
    @abstractmethod
    def query(
        self, table_name: str, time_interval: TimeInterval, filters: dict[str, object]
    ) -> pd.DataFrame:
        raise NotImplementedError



import pyspark.sql as ps

import datamazing.pandas as pdz


def assert_frame_equal(
    left: ps.DataFrame, right: ps.DataFrame, check_order: bool = False
):
    pdz.testing.assert_frame_equal(
        left=left.toPandas(),
        right=right.toPandas(),
        check_order=check_order,
    )



try:
    import pandas
except ImportError:
    raise ImportError("Missing optional dependency `pandas`.")

from . import testing
from .datacollection import Database
from .io import read_csv
from .transformations.basic import (
    as_of_time,
    concat_by_name,
    earliest,
    fill_empty_periods,
    latest,
    shift_time,
    switch,
    unique_one,
    unpivot,
)
from .transformations.grouping import Grouper, GrouperResampler, group
from .transformations.merging import align, merge, merge_many
from .transformations.reindexing import Reindexer, reindex
from .transformations.resampling import Resampler, resample
from .types import TimeInterval, TimeTravel



try:
    import pyspark
except ImportError:
    raise ImportError("Missing optional dependency `pyspark`.")

from . import testing
from .functions.time import ceil_time, floor_time, time_between
from .transformations.grouping import Grouper, group



import typing

import pandas as pd
import pyspark.sql as ps
import pyspark.sql.functions as f
from pyspark.sql.functions import date_format

from datamazing._conform import _concat, _list


class GrouperResampler:
    def __init__(
        self,
        gb: "Grouper",
        on: str,
        resolution: pd.Timedelta,
        edge: str,
    ):
        self.gb = gb
        self.on = on
        self.resolution = resolution
        self.edge = edge

    def agg(self, method: str):
        if self.edge == "right":
            round_func = f.ceil
        elif self.edge == "left":
            round_func = f.floor
        else:
            raise ValueError(f"Unsupported value {self.edge} for `edge`")

        resolution_sec = self.resolution.total_seconds()
        rounded_time = (
            round_func(f.unix_timestamp(self.on) / resolution_sec) * resolution_sec
        ).cast("timestamp")

        df = self.gb.df.withColumn(
            "__rounded_time", date_format(rounded_time, "yyyy-MM-dd HH:mm:ss")
        ).drop(self.on)

        df = group(df, by=_concat(self.gb.by, "__rounded_time")).agg(method)

        df = df.withColumn(self.on, f.col("__rounded_time").cast("timestamp"))

        df = df.drop("__rounded_time")

        return df


class Grouper:
    def __init__(self, df: ps.DataFrame, by: list[str]):
        self.df = df
        self.by = by

    def agg(self, method: str):
        remaining = set(self.df.columns).difference(_list(self.by))
        agg_func = getattr(f, method)

        return self.df.groupBy(self.by).agg(
            *[agg_func(col).alias(col) for col in remaining]
        )

    def latest(self, on: str):
        version_window = ps.Window.partitionBy(self.by).orderBy(f.desc(on))
        return (
            self.df.withColumn("__version", f.row_number().over(version_window))
            .filter(f.col("__version") == 1)
            .drop("__version")
        )

    def earliest(self, on: str):
        version_window = ps.Window.partitionBy(self.by).orderBy(f.asc(on))
        return (
            self.df.withColumn("__version", f.row_number().over(version_window))
            .filter(f.col("__version") == 1)
            .drop("__version")
        )

    def resample(self, on: str, resolution: pd.Timedelta, edge: str = "left"):
        return GrouperResampler(self, on, resolution, edge)

    def pivot(self, on: list[str], values: typing.Optional[list[tuple[str]]] = None):
        """
        Pivot table. Non-existing combinations will be filled
        with NaNs.

        Args:
            on (list[str]): Columns which to pivot
            values (list[tuple[str]], optional): Enforce
                the existence of columns with these names
                after pivoting. Defaults to None, in which
                case the values will be inferred from the
                pivoting column.
        """
        if values:
            # if values is a list of tuples, concatenate these
            # so that they match the resulting pivoted columns
            values = [
                "_".join([str(item) for item in _list(value)]) for value in values
            ]

        remaining = set(self.df.columns).difference(_list(on), _list(self.by))

        df = (
            self.df.withColumn("__on", f.concat_ws("_", *_list(on)))
            .groupBy(self.by)
            .pivot("__on", values=values)
            .agg(*[f.first(col).alias(col) for col in remaining])
        )

        return df


def group(df: ps.DataFrame, by: list[str]):
    return Grouper(df, by)



from dataclasses import dataclass, field

import pandas as pd


@dataclass
class TimeTravel:
    """
    Class representing a time travel either
    - absolute: How did data look for time T at a point in time T0
    - relative: How did data look for time T at a horizon of H prior
        to the time block B containing T.

    Example:
    Consider a versioned time series (slowly changing dimension, type 2) with
    columns `time`, `value`, `valid_from` and `valid_to`. Time travelling
    absolutely to the point in time T0, means considering the time series
    where `valid_from <= T0 < valid_to`. Time travelling to relatively with
    horizon H and block B, menas consider the time series where
    `valid_from <= floor(time_utc, block) - horizon < valid_to`.

    Args:
        [absolute time travel]
            as_of_time (pd.Timestamp): Point in time to time travel to.
        [relative time travel]
            horizon (pd.Timedelta): Period to look into future.
            block (pd.Timedelta): Period used to partition the
                timeline into discrete blocks. The horizon
                will be relative to the start of each block.
    """

    as_of_time: pd.Timestamp = None
    horizon: pd.Timedelta = None
    block: pd.Timedelta = None
    tense: str = field(init=False)

    def __post_init__(self):
        if self.as_of_time is not None and (
            self.horizon is not None or self.block is not None
        ):
            raise ValueError(
                "Can only specify one of `as_of_time` and `horizon`/`block`"
            )

        if self.as_of_time is None and (self.horizon is None or self.block is None):
            raise ValueError("Must specify one of `as_of_time` and `horizon`/`block`")

        if self.as_of_time is not None:
            self.tense = "absolute"
        else:
            self.tense = "relative"

    def __hash__(self):
        return hash((self.as_of_time, self.horizon, self.block))



import typing

import pandas as pd
import pyspark.sql as ps

import datamazing.pandas as pdz
from datamazing.pandas.datacollection import TimeInterval
from datamazing.pandas.testing.data import get_filepath  # noqa: F401


def make_df(data: list[dict]):
    spark = ps.SparkSession.getActiveSession()
    pdf = pdz.testing.make_df(data)
    return spark.createDataFrame(pdf)


def read_df(
    filename: str,
    subfolder: str = "data",
) -> ps.DataFrame:
    """
    Read pyspark DataFrame from test data.
    Datetimes and timedeltas are inferred automatically.

    Args:
        filename (str): Name of CSV file with test data
        subfolder (str, optional): Subfolder relative to test being
            run currently (taken from  the environment variable PYTEST_CURRENT_TEST),
            from where to read the test data. Defaults to "data".

    Example:
    >>> read_df(filename="data.df.csv")
    """
    spark = ps.SparkSession.getActiveSession()

    pdf = pdz.testing.read_df(filename, subfolder)
    return spark.createDataFrame(pdf)


class TestDatabase:
    def __init__(
        self, file_pattern: str, subfolder: str = "data", time_column: str = "time_utc"
    ):
        """Read Database from test data.
        Datetimes and timedeltas are inferred automatically.

        Args:
            file_pattern (str): Pattern of CSV files with test data
                using placeholder `{table_name}`
            subfolder (str, optional): Subfolder relative to test being
                run currently (taken from  the environment variable
                PYTEST_CURRENT_TEST), from where to read the test data.
                Defaults to "data".

        Example:
        >>> TestData(file_pattern="{table_name}.df.csv")
        """
        self.pdb = pdz.testing.TestDatabase(file_pattern, subfolder, time_column)

    def query(
        self, table_name: str, time_interval: typing.Optional[TimeInterval] = None
    ) -> pd.DataFrame:
        spark = ps.SparkSession.getActiveSession()

        pdf = self.pdb.query(table_name, time_interval)
        return spark.createDataFrame(pdf)



import pandas as pd


def assert_frame_equal(
    left: pd.DataFrame,
    right: pd.DataFrame,
    check_order: bool = False,
):
    if check_order:
        pd.testing.assert_frame_equal(left, right)
    else:
        columns = list(left.columns)
        pd.testing.assert_frame_equal(
            left=left.sort_values(by=columns).reset_index(drop=True),
            right=right.sort_values(by=columns).reset_index(drop=True),
            check_like=True,
        )


def assert_series_equal(
    left: pd.Series,
    right: pd.Series,
    check_order: bool = False,
):
    if check_order:
        pd.testing.assert_series_equal(left, right)
    else:
        pd.testing.assert_series_equal(
            left.sort_index(),
            right.sort_index(),
        )



import pandas as pd


class Reindexer:
    def __init__(
        self,
        df: pd.DataFrame,
        on: str,
        at: pd.Series | pd.Timestamp,
    ):
        self.df = df
        self.on = on
        self.at = at

    def interpolate(self, method: str = "linear"):
        """Interpolate time series.
        For example, if the input is a time series D with
        daily resolution, then we can interpolate, using
        a linear function, to another set of timestamps,
        and produce a new time series T. In other words:

            - Input:    D(day)
            - Output:   T(t) = interpolate(t between day+0 and day+1)

        Args:
            method (str): Interpolation method ("linear", "ffill" or "bfill")
        """
        # interpolating at a single point
        # should return a Series, instead
        # of a DataFrame
        if isinstance(self.at, pd.Timestamp):
            single_point = True
        else:
            single_point = False

        self.at = pd.Series(self.at)

        # only allow certain interpolation methods
        # to avoid untested behavior
        supported_methods = ["linear", "ffill", "bfill"]
        if method not in supported_methods:
            raise ValueError(f"Method must be one of {supported_methods}")

        # in pandas, the interpolation method `linear`
        # ignores the time values. We need to use `time`
        if method == "linear":
            method = "time"

        # pandas doesn't handle empty dataframes very well
        if self.df.empty:
            return self.df

        df = self.df.sort_values(by=self.on)

        union_index = pd.concat([self.at, df[self.on]]).drop_duplicates().sort_values()

        df = (
            df.set_index(self.on)
            .reindex(union_index)
            .interpolate(method)
            .reindex(self.at)
            .rename_axis(self.on)
            .reset_index()
        )

        if single_point:
            s = df.iloc[0]
            s.name = None
            return s

        return df


def reindex(df: pd.DataFrame, on: str, at: pd.Series | pd.Timestamp):
    return Reindexer(df, on, at)



import pandas as pd
import pyspark.sql.functions as f
from pyspark.sql.types import FloatType, TimestampType


def time_between(t1: TimestampType, t2: TimestampType, unit: pd.Timedelta) -> FloatType:
    seconds = unit.total_seconds()
    return f.try_subtract(t2, t1).cast("long").cast("double") / seconds


def floor_time(t: TimestampType, nearest: pd.Timedelta) -> TimestampType:
    seconds = nearest.total_seconds()
    return f.from_unixtime(f.floor(f.unix_timestamp(t) / seconds) * seconds).cast(
        "timestamp"
    )


def ceil_time(t: TimestampType, nearest: pd.Timedelta) -> TimestampType:
    seconds = nearest.total_seconds()
    return f.from_unixtime(f.ceil(f.unix_timestamp(t) / seconds) * seconds).cast(
        "timestamp"
    )



from os import PathLike

import pandas as pd

ISO_FORMAT = "%Y-%m-%dT%H:%M:%S%z"
ISO_FORMAT_MILLISECONDS = "%Y-%m-%dT%H:%M:%S.%f%z"


def infer_iso_datetime(values: pd.Series) -> pd.Series:
    try:
        converted_values = pd.to_datetime(values, format=ISO_FORMAT_MILLISECONDS)
    except (ValueError, TypeError):
        try:
            converted_values = pd.to_datetime(values, format=ISO_FORMAT)
        except (ValueError, TypeError):
            # if not possible to parse as datetime, return original values
            return values
    return converted_values


def infer_iso_timedelta(values: pd.Series) -> pd.Series:
    try:
        converted_values = pd.to_timedelta(values)
    except (ValueError, TypeError):
        # if not possible to parse as time delta, return original values
        return values
    try:
        values_isoformat = converted_values.apply(pd.Timedelta.isoformat)
    except (TypeError, AttributeError):
        return values
    if not (values_isoformat == values).all():
        # if original values is not in ISO 8601 format, return original values
        return values
    return converted_values


def read_csv(filepath: PathLike) -> pd.DataFrame:
    """
    Read CSV into DataFrame. All types are inferred.
    Datetimes and timedeltas in ISO 8601 format are inferred automatically.

    One can specify the type in the column name instead of relying on the inferred type.
    This is done using the following syntax: <column_name>:<type>.

    Some conversions are invalid and the will cause the reading to fail.
    Example of invalid conversions:
    - converting a string to a float

    Args:
        filepath (str): Filepath of the CSV file
    """
    df = pd.read_csv(filepath, keep_default_na=False, na_values=["nan"])

    # Auto cast types based on type hint in column name
    # type hints are on the form <column_name>:<type_hint>
    renames = {}
    for column_name in df.columns:
        if ":" in column_name:
            name, type_hint = column_name.split(":")
            df[column_name] = df[column_name].astype(type_hint)
            renames[column_name] = name
        else:
            # try converting ISO 8601 strings to pd.Timestamp and pd.Timedelta
            df[column_name] = infer_iso_datetime(df[column_name])
            df[column_name] = infer_iso_timedelta(df[column_name])

    df = df.rename(columns=renames)

    # convert all nan strings to None
    df = df.where(df != "nan", None)

    return df



import typing

import pandas as pd

from datamazing._conform import _concat, _list
from datamazing.pandas.transformations import resampling


class GrouperResampler:
    def __init__(
        self,
        gb: "Grouper",
        on: str,
        resolution: pd.Timedelta,
    ):
        self.gb = gb
        self.on = on
        self.resolution = resolution

    def agg(self, method: str | dict, edge: str = "left"):
        """Aggregate (downsample) time series.
        For example, if the input is a time series H with
        hourly resolution, then we can aggregate, using
        the mean, to daily resolution, and produce a new
        time series D. In other words:

            - Input:    H(hour)
            - Output:   D(day) = mean( H(hour) for hour in day)

        Args:
            method (str | dict): Name of method (e.g. "sum") or
                dictionary of method(s) to use for each column
                (e.g. {"col1": "mean", "col2": ["min", "max"]})
            edge (str, optional): Which side of the interval
                to use as label ("left" or "right").
                Defaults to "left".
        """
        df = (
            self.gb.df.sort_values(by=self.on)
            .set_index(self.on)
            .groupby(self.gb.by, dropna=False, observed=True)
            .resample(rule=self.resolution, closed=edge, label=edge)
            .aggregate(method, numeric_only=True)
        )

        # depending on the resampling aggregation
        # method, pandas will include the group-by
        # columns in both the index and the columns
        df = df.drop(columns=self.gb.by, errors="ignore")

        df = df.reset_index()

        if isinstance(df.columns, pd.MultiIndex):
            df.columns = ["_".join(col).strip("_") for col in df.columns]

        return df

    def interpolate(self, method: str = "linear"):
        """Interpolate (upsample) time series.
        For example, if the input is a time series D with
        daily resolution, then we can interpolate, using
        a linear function, to hourly resolution, and
        produce a new time series H. In other words:

            - Input:    D(day)
            - Output:   H(hour) = interpolate(hour between day+0 and day+1)

        Args:
            method (str): Interpolation method ("linear", "ffill" or "bfill")

        Raises:
            ValueError: Currently, interpolation can only be done
                when the original time series has a resolution
                which is an integer factor of the upsample resolution.
                Thus, one can go from 1 hour-resolution to 5 minute
                -resolution, but not the other way around. This
                restriction is purely a pandas issue, and can be
                fixed if desired.
        """
        # only allow certain interpolation methods
        # to avoid untested behavior
        supported_methods = ["linear", "ffill", "bfill"]
        if method not in supported_methods:
            raise ValueError(f"Method must be one of {supported_methods}")

        # in pandas, the aggregation method `linear`
        # doesn't exist. We need to use `interpolate`
        if method == "linear":
            method = "interpolate"

        df = self.gb.df.sort_values(by=self.on)

        diff_factors = (
            df.groupby(self.gb.by, observed=True)[self.on].diff().dropna()
            / self.resolution
        )

        if ((diff_factors % 1) != 0).any():
            raise ValueError(
                f"Interpolation only possible "
                f"when the original time resolution "
                f"is an integer factor of the upsample "
                f"resolution '{self.resolution}'."
            )

        start_time = df[self.on].min()
        end_time = df[self.on].max()

        df = (
            df.set_index(self.on)
            .groupby(self.gb.by, dropna=False, observed=True)
            .resample(rule=self.resolution)
            .aggregate(method)
        )

        # depending on the resampling aggregation
        # method, pandas will include the group-by
        # columns in both the index and the columns
        df = df.drop(columns=self.gb.by, errors="ignore")

        df = df.reset_index()

        # after resampling, pandas might leave
        # timestamps outside of the original interval
        if not df.empty:
            df = df[df[self.on].between(start_time, end_time)]

        return df

    def granulate(self, block_size: pd.Timedelta, edge: str = "left"):
        """Fine-grain (upsample) time series.
        For example, if the input is a time series D with
        daily resolution, then we can granulate to hourly
        resolution, and produce a new time series H.
        In other words:

            - Input:    D(day)
            - Output:   H(hour) = D(day containing hour)


        Args:
            block_size: The block size in which the time is granulated.
            edge (str, optional): Which side of the interval
                to use as label ("left" or "right").
                Defaults to "left".
        """
        # pandas doesn't handle empty dataframes very well
        if self.gb.df.empty:
            return self.gb.df

        df = self.gb.df.groupby(
            self.gb.by, dropna=False, group_keys=False, observed=True
        ).apply(
            lambda group: resampling.resample(
                group, self.on, self.resolution
            ).granulate(block_size, edge)
        )

        return df


class Grouper:
    def __init__(self, df: pd.DataFrame, by: list[str]):
        self.df = df
        self.by = by

    def agg(self, method: str | dict):
        """Aggregate table.

        Args:
            method (str | dict): Name of method (e.g. "sum") or
                dictionary of method(s) to use for each column
                (e.g. {"col1": "mean", "col2": ["min", "max"]})
        """
        if self.df.empty:
            aggregate_options = {}
        elif method in ["sum", "mean", "std", "var"]:
            aggregate_options = {"numeric_only": "True"}
        else:
            aggregate_options = {}
        df = (
            self.df.set_index(self.by)
            .groupby(self.by, dropna=False, observed=True)
            .aggregate(method, **aggregate_options)
            .reset_index()
        )
        if isinstance(df.columns, pd.MultiIndex):
            df.columns = ["_".join(col).strip("_") for col in df.columns]

        return df

    def resample(self, on: str, resolution: pd.Timedelta):
        return GrouperResampler(self, on, resolution)

    def pivot(self, on: list[str], values: typing.Optional[list[tuple[str]]] = None):
        """
        Pivot table. Non-existing combinations will be filled
        with NaNs.

        Args:
            on (list[str]): Columns which to pivot
            values (list[tuple[str]], optional): Enforce
                the existence of columns with these names
                after pivoting. Defaults to None, in which
                case the values will be inferred from the
                pivoting column.
        """

        df = self.df.set_index(_concat(self.by, on))

        if values:
            by_vals = df.index.to_frame(index=False)[_list(self.by)].drop_duplicates()
            on_vals = pd.DataFrame(values, columns=_list(on))
            cross_vals = by_vals.merge(on_vals, how="cross")
            df = df.reindex(pd.MultiIndex.from_frame(cross_vals))

        df = df.unstack(on)

        # concatenate multiindex columns to single index columns
        concat_cols = []
        suffix = len(df.columns.levels[0]) > 1
        for col in df.columns:
            concat_col = "_".join([str(item) for item in col[1:]])
            if suffix:
                # if more than one remaning columns, suffix with that
                concat_col = concat_col + "_" + str(col[0])
            concat_col = concat_col.strip("_")
            concat_cols.append(concat_col)
        df.columns = concat_cols

        return df.reset_index()

    def latest(self, on: str):
        return (
            self.df.set_index(_concat(self.by, on))
            .sort_index(level=on)
            .groupby(self.by, dropna=False, observed=True)
            .tail(1)
            .reset_index()
        )

    def earliest(self, on: str):
        return (
            self.df.set_index(_concat(self.by, on))
            .sort_index(level=on)
            .groupby(self.by, dropna=False, observed=True)
            .head(1)
            .reset_index()
        )


def group(df: pd.DataFrame, by: list[str]):
    return Grouper(df, by)
